# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "MOEXAlgo"
copyright = "2025, MOEX Algopack team"
author = "MOEX Algopack team"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
]

templates_path = ["_templates"]
exclude_patterns = []

language = "ru"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
    ".txt": "markdown",
}



import asyncio
import typing as t
from time import sleep

import httpx

from moexalgo.session import BaseClient
from moexalgo.utils import json


class Client(BaseClient):
    """
    Клиент для работы с API AlgopackHistorical Data
    """

    def get(self, path: str, **params: t.Any) -> t.Any:

        def _parse_response(resp: httpx.Response) -> t.Any:
            if resp.url.raw_path == b"/error?message=Authentication+required+to+download+files":
                resp = httpx.Response(401, request=resp.request)
            if not resp.is_success:
                resp.raise_for_status()
            if resp.headers["content-type"].startswith("application/json"):
                if data := json.loads(resp.text):
                    return data
            return resp.content

        def _sync_get(timeout: float = 0):
            if timeout:
                sleep(timeout)
            return _parse_response(self.httpx_cli.get(path, params=params))

        async def _async_get(timeout: float = 0):
            if timeout:
                await asyncio.sleep(timeout)
            return _parse_response(await self.httpx_cli.get(path, params=params))

        return _sync_get() if self.sync else _async_get()



import re
from datetime import datetime, timedelta, date, time


def normalize_period(period: int | str) -> tuple[int, int, str]:
    if isinstance(period, str):
        if found := re.match("([0-9]+)([a-zA-Z]+)", period):
            number, symbol = found.groups()
            number = int(number)
            if period == found.group():
                if symbol == "min":
                    if number in (5, 10, 15):
                        return 1, int(number), f"{number}min"
                    elif number in (20, 30):
                        return 10, int(number), f"{number}min"
                elif symbol in ("h", "H"):
                    if number in (1, 2, 3, 6, 12):
                        return 60, number * 60, f"{number}h"
                elif symbol in ("d", "D"):
                    if number in (1, 5, 10):
                        return 24, number * 60 * 24, f"{number}d"
                elif symbol in ("w", "W"):
                    if number in (1, 2, 4):
                        return 7, number * 60 * 24 * 7, f"{number}d"

    raise ValueError(
        f"Wrong period value: {period}. "
        ""
        "Accepted period values: '1min', '5min', '10min', '15min', '20min', '30min', "
        "'1h', '2h', '3h', '6h', '12h', '1D', '5D', '10D', '1W', '2W', '4W', '1M'"
    )


def _intervals_gen(from_date: date | datetime, period: int):
    if isinstance(from_date, datetime):
        from_date = from_date.date()
    begin = datetime.combine(from_date, time(0))
    end = begin + timedelta(minutes=period)
    while begin.year <= date.today().year:
        yield begin, end - timedelta(seconds=1)
        begin = end
        end = begin + timedelta(minutes=period)


def candles(data, minutes: int):
    accum = list()
    intervals = None

    def make_candle(start, finish):
        candle = dict(
            begin=start.isoformat(), end=finish.isoformat(), open=0, high=0, low=0, close=0, volume=0, value=0
        )
        if accum:
            candle.update(
                {
                    "open": accum[0]["open"],
                    "close": accum[-1]["close"],
                    "high": max([item["high"] for item in accum]),
                    "low": min([item["low"] for item in accum]),
                    "volume": sum([item["volume"] for item in accum]),
                    "value": round(sum([item["value"] for item in accum])),
                }
            )
            accum.clear()
        return candle

    counter = 0
    start = finish = None
    for candle in data:
        begin = datetime.fromisoformat(candle["begin"])
        end = datetime.fromisoformat(candle["end"])
        if intervals is None:
            intervals = _intervals_gen(begin, minutes)
            start, finish = next(intervals)
            while not end <= finish:
                start, finish = next(intervals)
        if begin > finish:
            if accum or counter:
                yield make_candle(start, finish)
                counter += 1
            start, finish = next(intervals)
        if start <= begin and end <= finish:
            accum.append(candle)
    yield make_candle(start, finish)



import re
import statistics
import typing as t
from datetime import datetime, timedelta, date, time

import moexalgo

try:
    import pandas as pd

    type Data = list[dict[str, t.Any]] | pd.DataFrame
except ImportError:
    type Data = list[dict[str, t.Any]]

type Metrics = t.Literal["tradestats", "orderstats", "obstats"]


def _normalize(data: list[dict[str, t.Any]]) -> list[dict[str, t.Any]]:
    def normalize(item: dict[str, t.Any]):
        if ticker := item.pop("ticker", item.pop("secid", None)):
            end = item.pop("end", None)
            begin = item.pop("begin", None)
            tradedate = item.pop("tradedate", None)
            tradetime = item.pop("tradetime", None)
            if tradedate and tradetime:
                tradedate = date.fromisoformat(tradedate) if isinstance(tradedate, str) else tradedate
                tradetime = time.fromisoformat(tradetime) if isinstance(tradetime, str) else tradetime
                end = datetime.combine(tradedate, tradetime)
                begin = end - timedelta(minutes=5)
            if end and begin:
                begin = datetime.fromisoformat(begin) if isinstance(begin, str) else begin
                end = datetime.fromisoformat(end) if isinstance(end, str) else end
                return dict(ticker=ticker, begin=begin, end=end, **item)
        raise ValueError(f"Wrong item: {item}")

    return list(map(normalize, data))


def _resolve_and_normalize(data: list[dict[str, t.Any]]) -> tuple[Metrics | None, list[dict[str, t.Any]]]:
    metrics = None
    if data:
        match data[0]:
            case {"trades": trades, "vol": vol}:
                if trades and vol:
                    metrics = "tradestats"
            case {"put_orders_b": put_orders, "cancel_orders_b": cancel_orders}:
                if put_orders or cancel_orders:
                    metrics = "orderstats"
            case {"levels_b": levels_b, "levels_s": levels_s}:
                if levels_b or levels_s:
                    metrics = "obstats"
        if metrics:
            data = _normalize(data)
    return metrics, data


def normalize_period(period: int | str) -> tuple[int, str]:
    if isinstance(period, str):
        if found := re.match("([0-9]+)([a-zA-Z]+)", period):
            number, symbol = found.groups()
            number = int(number)
            if period == found.group():
                if symbol == "min":
                    if number in (5, 10, 15, 20, 30):
                        return int(number), f"{number}min"
                elif symbol in ("h", "H"):
                    if number in (1, 2, 3, 6, 12):
                        return number * 60, f"{number}h"
                elif symbol in ("d", "D"):
                    if number in (1, 5, 10):
                        return number * 60 * 24, f"{number}d"
                elif symbol in ("w", "W"):
                    if number in (1, 2, 4):
                        return number * 60 * 24 * 7, f"{number}d"
                elif symbol == "M" and number == 1:
                    return None, "1M"

    raise ValueError(
        f"Wrong period value: {period}. "
        ""
        "Accepted period values: '5min', '10min', '15min', '20min', '30min', "
        "'1h', '2h', '3h', '6h', '12h', '1D', '5D', '10D', '1W', '2W', '4W', '1M'"
    )


def normalize_int(value):
    try:
        value = int(float(value))
    except (ValueError, TypeError):
        return None
    return value


def normalize_float(value, decimals=None):
    try:
        value = float(value)
        if decimals is not None:
            value = round(value, decimals)
    except (ValueError, TypeError):
        return None
    return value


def save_(func, items, name):
    items = [item[name] for item in items if item[name] is not None] if items else []
    return func(items) if items else None


def _tradestats_calculator(items: list[dict[str, t.Any]], lotsize: int, decimals: int) -> dict[str, t.Any]:
    TO_MEAN = ["pr_std"]
    TO_SUM = ["vol", "val", "trades", "trades_b", "trades_s", "vol_b", "vol_s", "val_b", "val_s"]
    items = [
        dict(
            (
                k,
                (
                    normalize_int(v)
                    if k
                    in (
                        "trades",
                        "trades_b",
                        "trades_s",
                        "vol",
                        "vol_b",
                        "vol_s",
                        "sec_pr_open",
                        "sec_pr_high",
                        "sec_pr_low",
                        "sec_pr_close",
                    )
                    else (
                        normalize_float(v)
                        if k
                        in (
                            "pr_std",
                            "val",
                            "val_b",
                            "val_s",
                            "pr_open",
                            "pr_high",
                            "pr_low",
                            "pr_close",
                            "oi_open",
                            "oi_high",
                            "oi_low",
                            "oi_close",
                        )
                        else v
                    )
                ),
            )
            for k, v in item.items()
        )
        for item in items
    ]

    result = None
    for N in range(len(items)):
        tradestat = items[N]
        offset = tradestat.pop("offset")
        if result is None:
            result = tradestat
            result["sec_pr_open"] = offset + tradestat["sec_pr_open"]
            result["sec_pr_high"] = offset + tradestat["sec_pr_high"]
            result["sec_pr_low"] = offset + tradestat["sec_pr_low"]
            result["sec_pr_close"] = offset + tradestat["sec_pr_close"]

        if tradestat["pr_high"] > result["pr_high"]:
            result["pr_high"] = tradestat["pr_high"]
            result["sec_pr_high"] = offset + tradestat["sec_pr_high"]
        if tradestat["pr_low"] < result["pr_low"]:
            result["pr_low"] = offset + tradestat["pr_low"]
            result["sec_pr_low"] = offset + tradestat["sec_pr_low"]
        result["pr_close"] = tradestat["pr_close"]
        result["sec_pr_close"] = offset + tradestat["sec_pr_close"]
        if "im" in tradestat:
            result["im"] = tradestat["im"]

        if "oi_open" in tradestat:
            if tradestat["oi_high"] > result["oi_high"]:
                result["oi_high"] = tradestat["oi_high"]
            if tradestat["oi_low"] < result["oi_low"]:
                result["oi_low"] = tradestat["oi_low"]
            result["oi_close"] = tradestat["oi_close"]

    for name in TO_SUM:
        if name in result:
            result[name] = save_(sum, items, name)
    for name in TO_MEAN:
        if name in result:
            try:
                result[name] = save_(statistics.mean, items, name)
            except statistics.StatisticsError:
                pass
    result["sec_pr_open"] = int(result["sec_pr_open"])
    result["sec_pr_high"] = int(result["sec_pr_high"])
    result["sec_pr_low"] = int(result["sec_pr_low"])
    result["sec_pr_close"] = int(result["sec_pr_close"])
    result["pr_std"] = round(result["pr_std"], 4) if result["pr_std"] else None
    result["pr_change"] = round(100 * (result["pr_close"] - result["pr_open"]) / result["pr_open"], 4)
    result["disb"] = round((result["vol_b"] - result["vol_s"]) / result["vol"], 2) if result["vol"] > 0 else None
    result["pr_vwap"] = round(result["val"] / (result["vol"] * lotsize), decimals) if result["vol"] else None
    result["pr_vwap_b"] = round(result["val_b"] / (result["vol_b"] * lotsize), decimals) if result["vol_b"] else None
    result["pr_vwap_s"] = round(result["val_s"] / (result["vol_s"] * lotsize), decimals) if result["vol_s"] else None

    for key, value in tuple(result.items()):
        if "val" in key:
            result[key] = normalize_float(value, 0)
    return result


def _orderstats_calculator(items: list[dict[str, t.Any]], lotsize: int, decimals: int) -> dict[str, t.Any]:
    TO_SUM = [
        "put_orders_b",
        "put_orders_s",
        "put_val_b",
        "put_val_s",
        "put_vol_b",
        "put_vol_s",
        "put_vol",
        "put_val",
        "put_orders",
        "cancel_orders_b",
        "cancel_orders_s",
        "cancel_val_b",
        "cancel_val_s",
        "cancel_vol_b",
        "cancel_vol_s",
        "cancel_vol",
        "cancel_val",
        "cancel_orders",
    ]

    items = [
        dict(
            (
                k,
                (
                    normalize_int(v)
                    if k
                    in (
                        "put_orders",
                        "put_vol",
                        "put_orders_b",
                        "put_orders_s",
                        "put_vol_b",
                        "put_vol_s",
                        "cancel_orders",
                        "cancel_vol",
                        "cancel_orders_b",
                        "cancel_orders_s",
                        "cancel_vol_b",
                        "cancel_vol_s",
                    )
                    else (
                        normalize_float(v)
                        if k in ("put_val", "put_val_b", "put_val_s", "cancel_val", "cancel_val_b", "cancel_val_s")
                        else v
                    )
                ),
            )
            for k, v in item.items()
        )
        for item in items
    ]

    result = None
    for N in range(len(items)):
        orderstat = items[N]
        if result is None:
            result = orderstat
            continue
    for name in TO_SUM:
        if name in result:
            result[name] = save_(sum, items, name)
    result["put_vwap_b"] = (
        round(result["put_val_b"] / (result["put_vol_b"] * lotsize), decimals) if result["put_vol_b"] else None
    )
    result["cancel_vwap_b"] = (
        round(result["cancel_val_b"] / (result["cancel_vol_b"] * lotsize), decimals)
        if result["cancel_vol_b"]
        else None
    )
    result["put_vwap_s"] = (
        round(result["put_val_s"] / (result["put_vol_s"] * lotsize), decimals) if result["put_vol_s"] else None
    )
    result["cancel_vwap_s"] = (
        round(result["cancel_val_s"] / (result["cancel_vol_s"] * lotsize), decimals)
        if result["cancel_vol_s"]
        else None
    )
    for key, value in tuple(result.items()):
        if "val" in key:
            result[key] = normalize_float(value, 0)
    return result


def _obstats_calculator(items: list[dict[str, t.Any]], lotsize: int, decimals: int) -> dict[str, t.Any]:
    TO_MEAN = [
        "spread_bbo",
        "spread_lv10",
        "spread_1mio",
        "spread_l1",
        "spread_l10",
        "spread_l2",
        "spread_l3",
        "spread_l5",
        "spread_l20",
        "levels_b",
        "levels_s",
        "imbalance_val",
        "imbalance_vol_bbo",
        "imbalance_val_bbo",
        "vwap_b_1mio",
        "vwap_s_1mio",
        "vol_b",
        "vol_s",
        "val_b",
        "val_s",
        "vol_b_l1",
        "vol_b_l10",
        "vol_b_l2",
        "vol_b_l3",
        "vol_b_l5",
        "vol_b_l20",
        "vol_s_l1",
        "vol_s_l10",
        "vol_s_l2",
        "vol_s_l20",
        "vol_s_l3",
        "vol_s_l5",
        "micro_price",
        "mid_price",
    ]

    items = [
        dict(
            (
                k,
                (
                    normalize_int(v)
                    if k
                    in (
                        "levels_b",
                        "levels_s",
                        "vol_b",
                        "vol_s",
                        "vol_b_l1",
                        "vol_b_l10",
                        "vol_b_l2",
                        "vol_b_l3",
                        "vol_b_l5",
                        "vol_b_l20",
                        "vol_s_l1",
                        "vol_s_l10",
                        "vol_s_l2",
                        "vol_s_l20",
                        "vol_s_l3",
                        "vol_s_l5",
                    )
                    else (
                        normalize_float(v)
                        if k
                        in (
                            "spread_bbo",
                            "spread_lv10",
                            "spread_1mio",
                            "spread_l1",
                            "spread_l10",
                            "spread_l2",
                            "spread_l3",
                            "spread_l5",
                            "spread_l20",
                            "imbalance_vol_bbo",
                            "imbalance_val_bbo",
                            "vwap_b_1mio",
                            "vwap_s_1mio",
                            "val_b",
                            "val_s",
                            "micro_price",
                            "mid_price",
                            "imbalance_val",
                        )
                        else v
                    )
                ),
            )
            for k, v in item.items()
        )
        for item in items
    ]

    result = None
    for N in range(len(items)):
        obstat = items[N]
        if result is None:
            result = obstat
            continue
    for name in TO_MEAN:
        if name in result:
            try:
                result[name] = save_(statistics.mean, items, name)
            except statistics.StatisticsError:
                pass
    for vwap, val, vol in [
        ("vwap_b", "val_b", "vol_b"),
        ("vwap_b_l1", "val_b_l1", "vol_b_l1"),
        ("vwap_b_l2", "val_b_l2", "vol_b_l2"),
        ("vwap_b_l3", "val_b_l3", "vol_b_l3"),
        ("vwap_b_l5", "val_b_l5", "vol_b_l5"),
        ("vwap_b_l10", "val_b_l10", "vol_b_l10"),
        ("vwap_b_l20", "val_b_l20", "vol_b_l20"),
        ("vwap_s", "val_s", "vol_s"),
        ("vwap_s_l1", "val_s_l1", "vol_s_l1"),
        ("vwap_s_l2", "val_s_l2", "vol_s_l2"),
        ("vwap_s_l3", "val_s_l3", "vol_s_l3"),
        ("vwap_s_l5", "val_s_l5", "vol_s_l5"),
        ("vwap_s_l10", "val_s_l10", "vol_s_l10"),
        ("vwap_s_l20", "val_s_l20", "vol_s_l20"),
    ]:
        if vol in result and val in result:
            result[vwap] = round(result[val] / (result[vol] * lotsize), decimals) if result[vol] else None

    for key, value in tuple(result.items()):
        if "spread" in key:
            result[key] = normalize_float(value, 1)
        elif "imbalance" in key:
            result[key] = normalize_float(value, 2)
        elif "val" in key:
            result[key] = normalize_float(value, 0)
        elif "vol" in key or "levels" in key:
            result[key] = normalize_int(value)
        elif "vwap" in key or "price" in key:
            result[key] = normalize_float(value, decimals)
    return result


class Resampler:

    def __init__(self, metrics: Metrics, ticker: str, from_date: date, period: int, *args):
        match metrics:
            case "tradestats":
                self._calculator = _tradestats_calculator
            case "orderstats":
                self._calculator = _orderstats_calculator
            case "obstats":
                self._calculator = _obstats_calculator
            case _:
                raise NotImplementedError(f"Resampler is not support '{metrics}' metrics")

        self._ticker = ticker
        self._accum = list()
        self._calculator_args = args
        self._intervals = self._intervals_gen(from_date, period)
        self._begin, self._end = next(self._intervals)

    def __call__(self, item: dict[str, t.Any] = None) -> dict[str, t.Any] | None:
        result = None
        if item is not None:
            begin = item.pop("begin")
            end = item.pop("end")
            while not (self._begin <= begin and end <= self._end):
                if result is None:
                    result = self._get_result()
                self._begin, self._end = next(self._intervals)
            item["offset"] = (begin - self._begin).total_seconds()
            self._accum.append(item)
        else:
            result = self._get_result()
        return result

    def _get_result(self):
        result = None
        if self._accum:
            result = dict(
                ticker=self._ticker,
                begin=self._begin,
                end=self._end,
                **self._calculator(self._accum, *self._calculator_args),
            )
            result.pop("offset", None)
            self._accum.clear()
        return result

    @staticmethod
    def _intervals_gen(from_date: date, period: int | None):
        if period is None:
            begin = datetime.combine(date(from_date.year, from_date.month, 1), time(0))
            end = datetime.combine(
                date(begin.year, begin.month + 1, 1) if begin.month < 12 else date(begin.year + 1, 1, 1), time(0)
            )
            while True:
                yield begin, end
                begin = datetime.combine(
                    date(begin.year, begin.month + 1, 1) if begin.month < 12 else date(begin.year + 1, 1, 1), time(0)
                )
                end = datetime.combine(
                    date(begin.year, begin.month + 1, 1) if begin.month < 12 else date(begin.year + 1, 1, 1), time(0)
                )
        else:
            begin = datetime.combine(from_date, time(0))
            end = begin + timedelta(minutes=period)
            while begin.year <= date.today().year:
                yield begin, end
                begin = end
                end = begin + timedelta(minutes=period)

    @classmethod
    def it(cls, metrics, data: list[dict[str, t.Any]], period: int):
        resamplers = dict()
        ticker_args = dict(
            (item["ticker"], (item.get("lotsize", item.get("lotvolume", -1)), item["decimals"]))
            for item in sum(
                list(
                    [a for a in item]
                    for item in [
                        market.tickers(native=True)
                        for market in [moexalgo.Market(market) for market in ["EQ", "FX", "FO"]]
                    ]
                ),
                start=[],
            )
        )
        for item in data:
            ticker = item.pop("ticker")
            if ticker in ticker_args:
                resampler = resamplers.setdefault(
                    ticker, cls(metrics, ticker, item["begin"].date(), period, *ticker_args[ticker])
                )
                if result := resampler(item):
                    yield result
        for resampler in resamplers.values():
            if result := resampler():
                yield result


def resample(data: Data, period: int | str) -> Data:
    if not isinstance(data, list):
        data = data.to_dict("records")
        used_fd = True
    else:
        used_fd = False
    period, _ = normalize_period(period)
    metrics, data = _resolve_and_normalize(data)
    if metrics is None:
        raise NotImplementedError("Resample is not implement for this data")
    result = list(Resampler.it(metrics, data, period))
    return pd.DataFrame(result) if used_fd else result



import sys
from datetime import date
from types import ModuleType

import moexalgo.engines.currency
import moexalgo.engines.futures
import moexalgo.engines.stock
from moexalgo import engines
from moexalgo.features.algopack import AlgopackMarketMixin, AlgopackTickerMixin
from moexalgo.features.common import CommonMarket, CommonTicker
from moexalgo.features.extramarket import ExtraMarketMixin
from moexalgo.session import Session
from moexalgo.utils import Desc, CandlePeriod


def Market(name: str, board: str = None) -> CommonMarket | ExtraMarketMixin | AlgopackMarketMixin:
    """
    Представление раздела биржевого рынка.

    Parameters
    ----------
    name : str
        Название рынка или его символическое имя.
    board : str, optional
        Идентификатор, указывающий на специфическую торговую площадку или сегмент рынка.
        Если не указано, автоматически определятся первичный идентификатор на основе общих правил.
    """
    RESOLVE_MAP = {
        (engines.currency, "selt", "CETS"): Desc(["selt", "currency", "fx"], dict(algopack="datashop/algopack/fx")),
        (engines.stock, "shares", "TQBR"): Desc(
            ["shares", "stocks", "equity", "eq"], dict(algopack="datashop/algopack/eq")
        ),
        (engines.stock, "index", "SNDX"): Desc(["index"]),
        (engines.stock, "bonds", "TQOB"): Desc(["bonds"]),
        (engines.futures, "forts", "RFUD"): Desc(
            ["futures", "forts", "fo"],
            dict(algopack="datashop/algopack/fo", futoi="analyticalproducts/futoi/securities"),
        ),
        (engines.futures, "options", "ROPD"): Desc(["options"]),
    }

    def resolve(name: str, boardid: str | None) -> tuple[ModuleType, str, str, dict[str, str]]:
        for (module, market, default_boardid), (aliases, features) in RESOLVE_MAP.items():
            if name in aliases:
                return module, market, boardid or default_boardid, features or {}
        raise LookupError(f"Unrecognized market' name '{name}'")

    module, name, boardid, features = resolve(name.lower(), board.upper() if board else None)
    return getattr(module, "Market")(name, boardid, features)


def Ticker(name: str, board: str = None) -> CommonTicker | AlgopackTickerMixin:
    """
    Представление конкретного финансового инструмента.

    Parameters
    ----------
    name : str
        Тикер финансового инструмента, например "GAZP" для акций Газпрома.
    board : str, optional
        Идентификатор, указывающий на специфическую торговую площадку или сегмент рынка.
        Если не указано, автоматически определятся первичный идентификатор на основе общих правил.
    """

    def resolve(secid: str, boardid: str | None) -> tuple[str, str, str, int, bool]:
        with Session() as client:
            data = client.get_objects(
                f"securities/{secid}",
                lambda data: [dict(zip(data["boards"]["columns"], row)) for row in data["boards"]["data"]],
            )
        if found := [
            item
            for item in data
            if item["secid"] == secid and (item["boardid"] == boardid if boardid is not None else item["is_primary"])
        ]:
            match found[0]:
                case {"boardid": boardid, "market": market, "decimals": decimals, "listed_till": listed_till}:
                    return market, boardid, secid, decimals, date.fromisoformat(listed_till) < date.today()
        raise LookupError(f"Unrecognized ticker' name '{secid}'")

    market, boardid, secid, decimals, delisted = resolve(name.upper(), board.upper() if board is not None else None)
    market = Market(market, boardid)
    return getattr(sys.modules[market.__module__], "Ticker")(market, boardid, secid, decimals, delisted)



from __future__ import annotations

import asyncio
from time import time, sleep
from typing import Callable, Optional, Union

import httpx

from moexalgo.utils import json

BASE_URL = "https://iss.moex.com/iss"
TOKEN_URL = "https://apim.moex.com/iss"
AUTH_URL = "https://passport.moex.com/authenticate"
AUTH_CERT = None
USE_HTTPS = True
TOKEN = None
_NEXT_REQUEST_AT = 0
_REQUEST_TIMEOUT = 0.1


class HasOptions:
    """
    Базовый класс для объектов с опциями.

    Attributes
    ----------
    options : dict
        Опционные параметры.
    """

    def __init__(self, auth_cert: str = None, base_url: str = None, timeout: int = 300, **options) -> None:
        """
        Parameters
        ----------
        auth_cert : str
            Сертификат авторизации.
        base_url : str
            Базовый URL.
        timeout : int
            Таймаут
        options : dict
            Опционные параметры.

        Returns
        -------
        return : None
        """
        global TOKEN
        if TOKEN:
            if "headers" not in options:
                options["headers"] = []
            if not [name for name, _ in options["headers"] if name == "Authorization"]:
                options["headers"].append(("Authorization", f"Bearer {TOKEN}"))
        if not base_url:
            base_url = TOKEN_URL if TOKEN else BASE_URL
        if base_url.startswith("http:") and USE_HTTPS:
            base_url = base_url.replace("http:", "https:")
        elif base_url.startswith("https:") and not USE_HTTPS:
            base_url = base_url.replace("https:", "http:")
            options["verify"] = False

        self.__options = dict(**options, base_url=base_url, timeout=timeout)
        if auth_cert:
            self.__options.setdefault("cookies", {})["MicexPassportCert"] = auth_cert

    @property
    def options(self) -> dict:
        """
        Опционные параметры использованные при создании сессии/клиента.

        Returns
        -------
        return : dict
            Опционные параметры.
        """
        return self.__options


class BaseClient(HasOptions):
    """Base API client"""

    def __init__(self, sync: bool = True, **options) -> None:
        """
        Parameters
        ----------
        sync : bool
            Синхронный режим работы.
        options : dict
            Опционные параметры.

        Returns
        -------
        return : None
        """
        assert sync, "AsyncClient not supported in this version"
        options.update(follow_redirects=True, headers=[("User-Agent", "python-httpx/moexalgo")])
        super().__init__(**options)
        self.httpx_cli = httpx.Client(**self.options) if sync else httpx.AsyncClient(**self.options)

    @property
    def sync(self) -> bool:
        """
        Синхронный режим работы клиента.

        Returns
        -------
        return : bool
            `True` если клиент синхронный, иначе `False`.
        """
        return isinstance(self.httpx_cli, httpx.Client)


class Client(BaseClient):
    """
    Клиент для работы с API.

    Attributes
    ----------
    sync : bool
        Синхронный режим работы.
    authorized : bool
        Авторизован ли клиент.
    httpx_cli : httpx.Client | httpx.AsyncClient
        Клиент для работы с HTTP.
    """

    @property
    def authorized(self) -> bool:
        """
        Авторизован ли клиент.

        Returns
        -------
        return : bool
            `True` если клиент авторизован, иначе `False`.
        """
        return bool(self.options.get("auth_cert"))

    def authorize(self, username: str, password: str) -> Optional[str]:
        """
        Авторизация клиента.

        Parameters
        ----------
        username : str
            Имя пользователя.
        password : str
            Пароль пользователя.

        Returns
        -------
        return : Optional[str]
            Сертификат авторизации или None, если авторизация не удалась.
        """

        def _process_response(resp: httpx.Response) -> Optional[str]:
            """
            Обработка ответа авторизации.

            Parameters
            ----------
            resp : httpx.Response
                Ответ авторизации.

            Returns
            -------
            return : Optional[str]
                Сертификат авторизации или None, если авторизация не удалась.
            """
            if resp.is_success:
                return resp.cookies.get("MicexPassportCert", self.options.get("cookies", {}).get("MicexPassportCert"))
            return None

        async def _async_authorize() -> Optional[str]:
            """
            Асинхронная авторизация.

            Returns
            -------
            return : Optional[str]
                Сертификат авторизации или None, если авторизация не удалась.
            """
            return _process_response(await self.httpx_cli.get(AUTH_URL, auth=(username, password)))

        def _sync_authorize() -> Optional[str]:
            """
            Синхронная авторизация.

            Returns
            -------
            return : Optional[str]
                Сертификат авторизации или None, если авторизация не удалась.
            """
            return _process_response(self.httpx_cli.get(AUTH_URL, auth=(username, password)))

        self.options["auth_cert"] = _sync_authorize() if self.sync else _async_authorize()
        return self.options["auth_cert"]

    def get_objects(self, path: str, deserializer: Callable[[dict], dict | list], **params) -> Union[dict, list]:
        """
        Получение объектов по переданному адресу.

        Parameters
        ----------
        path : str
            Путь к объектам.
        deserializer: callable
            Функция десериализации.
        params : dict
            Параметры запроса.

        Returns
        -------
        return : Union[dict, list]
            Объекты или список объектов.

        Raises
        ------
        httpx.HTTPStatusError
            Вызывается, если запрос завершился неудачно.
        ValueError
            Вызывается, если получен неверный ответ.
        """

        global _NEXT_REQUEST_AT

        def _parse_response(resp: httpx.Response) -> Union[dict, list]:
            """
            Парсинг ответа.

            Parameters
            ----------
            resp : httpx.Response
                Ответ запроса.

            Returns
            -------
            return : Union[dict, list]
                Объекты или список объектов.
            """
            if not resp.is_success:
                resp.raise_for_status()
            if not resp.headers.get("content-type", "").startswith("application/json"):
                resp.status_code = 403
                resp.raise_for_status()
            if data := json.loads(resp.text):
                return deserializer(data)
            raise ValueError("Received wrong data")

        async def _async_get_objects(timeout: float) -> Union[dict, list]:
            """
            Асинхронное получение объектов.

            Parameters
            ----------
            timeout : float
                Таймаут (время ожидания).

            Returns
            -------
            return : Union[dict, list]
                Объекты или список объектов.
            """
            if timeout:
                await asyncio.sleep(timeout)
            return _parse_response(await self.httpx_cli.get(url, params=params))

        def _sync_get_objects(timeout: float) -> Union[dict, list]:
            """
            Синхронное получение объектов.

            Parameters
            ----------
            timeout : float
                Таймаут (время ожидания).

            Returns
            -------
            return : Union[dict, list]
                Объекты или список объектов.
            """
            if timeout:
                sleep(timeout)
            return _parse_response(self.httpx_cli.get(url, params=params))

        path = [item for item in path.split("/") if item.strip()]
        url = "/".join(path) + ".json"

        timeout = _NEXT_REQUEST_AT - time()
        if self.authorized or timeout <= 0:
            timeout = 0.2
        _NEXT_REQUEST_AT = time() + _REQUEST_TIMEOUT + timeout

        return _sync_get_objects(timeout) if self.sync else _async_get_objects(timeout)

    @staticmethod
    def format_error(exc: httpx.HTTPStatusError) -> str:
        """
        Форматирование ошибки.

        Parameters
        ----------
        exc : httpx.HTTPStatusError
            Исключение HTTP.

        Returns
        -------
        return : str
            Сообщение об ошибке и рекомендации.
        """
        url = str(exc.request.url).replace(".json", "")
        main_msg = f"HTTP request to {url}: failed with code: {exc.response.status_code};"
        recommendations = "Please authenticate" if exc.response.status_code == 403 else "Please try again later"
        return " ".join([main_msg, recommendations])


class Session(HasOptions):
    """
    Сессия для работы с API.

    Attributes
    ----------
    _client : Client
        Клиент для работы с API.
    """

    def __init__(self, cs: HasOptions = None, client_cls=Client, **options) -> None:
        """
        Parameters
        ----------
        cs : HasOptions | None
            Сессия, из которой будут взяты опции, by default None.
        options : dict
            Опционные параметры сессии.

        Returns
        -------
        return : None
        """
        # assert isinstance(cs, HasOptions) or len(options)
        if cs is not None:
            options.update(cs.options)
        super().__init__(**options)
        self._client_cls = client_cls
        self._client = None

    def __enter__(self) -> Client:
        """
        Вход в сессию.

        Returns
        -------
        return : Client
            Клиент для работы с API.
        """
        self._client = self._client_cls(True, **self.options)
        self._client.httpx_cli.__enter__()
        return self._client

    def __exit__(self, *exc_info) -> bool:
        """
        Выход из сессии.

        Parameters
        ----------
        exc_info : tuple
            Информация об исключении.

        Returns
        -------
        return : bool
            `True`, если исключение обработано, иначе `False`.
        """
        return self._client.httpx_cli.__exit__(*exc_info)

    async def __aenter__(self) -> Client:
        """
        Асинхронный вход в сессию.

        Returns
        -------
        return : Client
            Клиент для работы с API.
        """
        self._client = Client(False, **self.options)
        await self._client.httpx_cli.__aenter__()
        return self._client

    async def __aexit__(self, *exc_info):
        return await self._client.httpx_cli.__aexit__(*exc_info)


def authorize(username: str, password: str) -> bool:
    """
    Авторизация сессии по умолчанию.

    Parameters
    ----------
    username : str
        Имя пользователя.
    password : str
        Пароль пользователя.

    Returns
    -------
    return : bool
        `True`, если авторизация прошла успешно, иначе `False`.
    """
    global AUTH_CERT
    with Session(auth_cert=AUTH_CERT) as client:
        if auth_cert := client.authorize(username, password):
            AUTH_CERT = auth_cert
            return True
        return False



import os

import pytest

from moexalgo import Market, session

session.TOKEN = os.environ["APIKEY"]


def test_market():
    eq = Market("EQ")

    assert len(eq.tickers()) > 50
    assert len(eq.marketdata()) > 50

    it = eq.tradestats(date="2024-01-05")
    assert isinstance(next(it), object)
    data = list(it)
    assert len(data) > 10000

    assert len(eq.trades()) > 10
    assert len(eq.candles()) > 10


if __name__ == "__main__":
    pytest.main()



from typing import Iterable, Any

from moexalgo.features.algopack import AlgopackTickerMixin, AlgopackMarketMixin
from moexalgo.features.common import CommonMarket, CommonTicker
from moexalgo.features.extramarket import ExtraMarketMixin
from moexalgo.features.futoi import FutOIMarketMixin, FutOITickerMixin
from moexalgo.utils import DataFrame

FORTS_TICKERS_DEFAULT_FIELDS = (
    "sectype",
    "assetcode",
    "shortname",  # Краткое наименование
    "lotvolume",  # Размер лота
    "decimals",  # Количество знаков после запятой
    "minstep",  # Минимальный шаг цены
    "initialmargin",
    "lasttradedate",
)

FORTS_MARKET_DEFAULT_FIELDS = (
    "bid",  # Лучшая цена покупки
    "offer",  # Лучшая цена продажи
    "biddeptht",  # Глубина стакана покупки
    "offerdeptht",  # Глубина стакана продажи
    "open",  # Цена открытия
    "high",  # Максимальная цена
    "low",  # Минимальная цена
    "last",  # Цена последней сделки
    "waprice",  # Средневзвешенная цена
    "lasttoprevprice",  # Изменение цены последней сделки к предыдущей
    "numtrades",  # Количество сделок
    "voltoday",  # Объем сделок за день
    "valtoday",  # Объем сделок за день в валюте
    "valtoday_usd",  # Объем сделок за день в долларах
    "openperiodprice",  # Цена открытия периода
    "closingauctionprice",  # Цена закрытия аукциона
    "closingauctionvolume",  # Объем закрытия аукциона
    "issuecapitalization",  # Капитализация
    "updatetime",  # Время обновления
    "systime",  # Время системы
)


class Market(CommonMarket, ExtraMarketMixin, AlgopackMarketMixin, FutOIMarketMixin):
    """
    Обобщенный класс `Market` для срочного рынка.
    """

    def __init__(self, *args, **kwargs):
        super().__init__("futures", *args, **kwargs)

    def tickers(self, *fields: str, native: bool = False) -> Iterable[dict[str, Any]] | DataFrame:
        match self.market:
            case "forts":
                fields = () if "*" in fields else (fields or FORTS_TICKERS_DEFAULT_FIELDS)
            case _:
                fields = () if "*" in fields else fields
        return super().tickers(*fields, native=native)

    def marketdata(self, *fields: str, native: bool = False) -> Iterable[dict[str, Any]] | DataFrame:
        match self.market:
            case "forts":
                fields = () if "*" in fields else (fields or FORTS_MARKET_DEFAULT_FIELDS)
            case _:
                fields = () if "*" in fields else fields
        return super().marketdata(*fields, native=native)

    def trades(
        self: CommonMarket,
        recno: int | None = None,
        *,
        native: bool = False,
    ) -> Iterable[dict[str, Any]] | DataFrame:
        """
        Возвращает последние сделки в обратной сортировке если `recno` не заданно.
        Если заданно сделки возвращаются от заданного `recno`.

        Parameters
        ----------
        recno :
            Номер порядка заключения сделок, если не задано.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        return super().trades(recno, native=native)


class Ticker(CommonTicker, AlgopackTickerMixin, FutOITickerMixin):
    """
    Обобщенный класс `Ticker` для срочного рынка.
    """

    def trades(
        self,
        recno: int | None = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> Iterable[dict[str, Any]] | DataFrame:
        """
        Возвращает сделки за последний день или начиная с заданного `recno`.

        Parameters
        ----------
        recno :
            Номер порядка заключения сделок, если не задано.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        return super().trades(recno, offset=offset, latest=latest, native=native)



import asyncio
import json
import typing as t
import uuid
from collections import deque
from contextlib import asynccontextmanager

import websockets
from stomp.utils import Frame, convert_frame, parse_frame


class Credentials(t.NamedTuple):
    """Auth credentials"""

    domain: str
    login: str
    passcode: str


class Subscription(t.AsyncIterator):

    def __init__(self, id: str):
        self.id = id
        self._queue = deque(maxlen=256)

    def _append(self, data):
        self._queue.append(data)

    async def __anext__(self):
        while not len(self._queue):
            await asyncio.sleep(0.1)
        result = self._queue.popleft()
        if isinstance(result, Exception):
            raise result
        return result


class ISSPlusSTOMP:
    """ISS+ STOMP Client"""

    structure: dict[str, t.Any] = dict()

    def __init__(self, wscp: websockets.WebSocketClientProtocol, cred: Credentials):
        self._wscp = wscp
        self._cred = cred
        self._task = None  # type: asyncio.Task | None
        self._pending = dict()  # type: dict[str, asyncio.Future | Subscription]

    def __await__(self):
        return self.run_forever().__await__()

    async def __aenter__(self):
        auth_frame = Frame("CONNECT", headers=self._cred._asdict())
        await self._wscp.send(b"".join(convert_frame(auth_frame)))
        async for message in self._wscp:
            frame = parse_frame(message)
            if frame.cmd == "CONNECTED":
                self.structure = json.loads(frame.body.decode("utf8").strip("\0"))["structure"]
                self._task = asyncio.create_task(self._listener(), name="Message listener")
                return self
            raise ConnectionRefusedError(f"STOMP authentication failed; {frame.headers['message']}")

    async def __aexit__(self, *exc_info):
        if self._task:
            if not self._task.done():
                self._task.cancel()
        self._task = None

    async def request(self, destination, selector):
        id = str(uuid.uuid4())
        req_frame = Frame("REQUEST", headers=dict(id=id, destination=destination, selector=selector))
        await self._wscp.send(b"".join(convert_frame(req_frame)))
        future = self._pending[id] = asyncio.Future()
        return await future

    async def subscribe(self, destination, selector):
        id = str(uuid.uuid4())
        subs_frame = Frame("SUBSCRIBE", headers=dict(id=id, receipt=id, destination=destination, selector=selector))
        await self._wscp.send(b"".join(convert_frame(subs_frame)))
        subscription = self._pending[id] = Subscription(id)
        return subscription

    async def unsubscribe(self, id: str):
        unsub_frame = Frame("UNSUBSCRIBE", headers=dict(id=id))
        await self._wscp.send(b"".join(convert_frame(unsub_frame)))
        subscription = self._pending.pop(id, None)
        subscription._append(StopAsyncIteration())

    async def run_forever(self):
        if self._task is None:
            raise ConnectionError("Client is not connected")
        await self._task

    async def close(self):
        if self._wscp is None:
            raise ConnectionError("Client is not connected")
        while self._pending:
            item = self._pending[tuple(self._pending.keys())[0]]
            if isinstance(item, Subscription):
                await self.unsubscribe(item.id)
            elif isinstance(item, asyncio.Future):
                item.cancel()
            else:
                assert False
        await self._wscp.close()

    async def _listener(self):
        try:
            async for message in self._wscp:
                frame = parse_frame(message)
                if request_id := frame.headers.get("request-id"):
                    if future := self._pending.pop(request_id, None):
                        if frame.cmd == "ERROR":
                            future.set_exception(
                                RuntimeError(f"Request {request_id} failed: {frame.headers['message']}")
                            )
                        else:
                            future.set_result(json.loads(frame.body.decode("utf8").strip("\0")))
                    else:
                        assert False, f"Cannot found pending for request: {request_id}"
                elif subscription_id := frame.headers.get("subscription", frame.headers.get("receipt-id")):
                    if subscription := self._pending.get(subscription_id, None):
                        if frame.cmd == "ERROR":
                            subscription._append(
                                RuntimeError(f"Subscription {subscription_id} failed: {frame.headers['message']}")
                            )
                            self._pending.pop(subscription_id, None)
                        else:
                            data = frame.body.decode("utf8").strip("\0")
                            if data:
                                subscription._append(json.loads(data))
                    else:
                        assert False, f"Cannot found pending for subscription: {subscription_id}"
                else:
                    assert False
        except asyncio.CancelledError:
            pass
        except Exception as exc:
            print(exc)


@asynccontextmanager
async def connect(url: str, credentials: Credentials):
    async with websockets.connect(url, subprotocols=["STOMP"]) as websocket:
        async with ISSPlusSTOMP(websocket, credentials) as iss_plus:
            yield iss_plus



# /// script
# dependencies = [
#   "moexalgo[issplus]",
# ]
# ///
import asyncio
import os

from moexalgo.beta import issplus


async def out_lasttrades(user, password):
    selector = 'TICKER="MXSE.TQBR.{ticker}" and LANGUAGE="en"'
    url = "wss://iss.moex.com/infocx/v3/websocket"
    async with issplus.connect(url, issplus.Credentials("passport", user, password)) as client:
        # подписка на последние сделки, MOEX
        destination = "MXSE.lasttrades"
        subscription = await client.subscribe(destination, selector.format(ticker="MOEX"))
        # Отмена подписки через 45 сек
        asyncio.get_running_loop().call_later(45.0, lambda: asyncio.create_task(client.unsubscribe(subscription.id)))
        # Получение данных
        columns = []
        async for data in subscription:
            if not columns:
                columns.extend(data["columns"])
                print("\t".join(columns))
            for rec in data["data"]:
                print("\t".join(map(str, [round(*cell) if isinstance(cell, list) else cell for cell in rec])))


if __name__ == "__main__":
    asyncio.run(
        out_lasttrades(
            os.environ.get("USER", "<Your account>"),
            os.environ.get("PASSWORD", "<Your password"),
        )
    )



from urllib.parse import urlparse, parse_qs

from moexalgo import session
from moexalgo.ahdata.client import Client
from moexalgo.ahdata.models import FileInfo, FileList, File, DownloadInfo
from moexalgo.session import HasOptions, Session

URL = "https://ahdata.ru"


def files(dataset: str = None) -> FileList:
    path = f"/api/files/{dataset}" if dataset else f"/api/files"
    with Session(HasOptions(base_url=URL), client_cls=Client) as client:
        data = client.get(path)
        return FileList(FileInfo(dataset=dataset, **item) if dataset else FileInfo(**item) for item in data["files"])


def download(dataset: str, period: str, use_dataframe=True):
    params = {"dataset_path": dataset, "year_month": period, "jwt_token": session.TOKEN}
    with Session(HasOptions(base_url=URL), client_cls=Client) as client:
        data = client.get("/download", **params)
        p = urlparse(data["download_url"])
        url = p.scheme + "://" + p.hostname + p.path
        params = parse_qs(p.query)
        data = client.get(url, **params)
        return File(data).df if use_dataframe else File(data)



import typing as t
from datetime import date

from moexalgo.features.common import CommonMarket, CommonTicker
from moexalgo.session import Session
from moexalgo.utils import DataFrame, result_adapter, normalize_data, prepare_from_till_dates, calc_offset_limit


class FutOIMarketMixin:
    """
    Миксин добавляющий метод `futoi` обобщенному классу `Market`.
    """

    def futoi(
        self: CommonMarket,
        *,
        date: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает метрики `FUTOI` по заданным параметрам.

        Parameters
        ----------
        date :
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self._get_path_for("futoi")
        date, _ = prepare_from_till_dates(date, date)
        options = dict[str, t.Any](date=date)
        if latest:
            options["latest"] = 1
        offset, limit = calc_offset_limit(offset, 10_000)
        options = dict(options, offset=offset, limit=limit)
        return result_adapter(fetch_futui(path, **options), native)


class FutOITickerMixin:
    """
    Миксин добавляющий метод `futoi` обобщенному классу `Ticker`.
    """

    def futoi(
        self: CommonTicker,
        *,
        start: str | date = None,
        end: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает метрики `FUTOI` по заданным параметрам.

        Parameters
        ----------
        start :
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end :
            Дата конца диапазона выдачи данных.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self.market._get_path_for("futoi")
        from_date, till_date = prepare_from_till_dates(start, end)
        options: dict[str, t.Any] = {"from": from_date, "till": till_date}
        if latest:
            options["latest"] = 1
        offset, limit = calc_offset_limit(offset, 10_000)
        options = dict(options, offset=offset, limit=limit)
        if not hasattr(self, "_tickers"):
            self._sectypes = dict(
                (item["secid"], item["sectype"]) for item in self.market.tickers("secid", "sectype", native=True)
            )
        sectype = self._sectypes.get(self.secid, None)
        if not sectype:
            raise KeyError(f"sectype for secid: {self.secid} not found")
        return result_adapter(fetch_futui(path, sectype, **options), native)


def fetch_futui(
    path: str,
    secid: str = None,
    limit: int = 10_000,
    offset: int = 0,
    **options: t.Any,
):
    with Session() as client:
        start = offset
        while True:
            options["start"] = start
            if data := normalize_data(
                client.get_objects(f"{path}/{secid}" if secid else f"{path}", lambda data: data["futoi"], **options)
            ):
                for item in data:
                    yield item
                    start += 1
                if (start - offset) < limit and limit > 0:
                    continue
            break



"""
Пример использования интерфейса ISS+ основанного на технологии websocket.
Для запуска этого примера необходимо установить библиотеку MOEXAlgo c дополнительными
зависимостями `pip install moexalgo[issplus]`.
Установятся модули для работы с websocket+STOMP.
"""
import asyncio

from moexalgo.beta import issplus


async def sample_websocket(url):
    async with issplus.connect(url, issplus.Credentials('DEMO', 'guest', 'guest')) as client:
        print('\n\n\n== META Info ===')
        print(client.structure)

        # режим "Вопрос/Ответ"
        print('\n\n\n== Request: SEARCH.ticker, pattern="SBER"  ===')
        result = await client.request(R'SEARCH.ticker', 'pattern="SBER"')
        print(result)

        # режим "Подписка"
        subs = await client.subscribe('MXSE.securities', 'TICKER="MXSE.TQBR.GAZP" and LANGUAGE="en"')
        print(f'\n\n\n== Subscribe ({subs.id}): MXSE.securities, TICKER="MXSE.TQBR.GAZP" and LANGUAGE="en"  ===')

        # Отмена подписки через 45 сек
        asyncio.get_running_loop().call_later(45.0, lambda: asyncio.create_task(client.unsubscribe(subs.id)))
        # Отключение клиента через 50 сек
        asyncio.get_running_loop().call_later(50.0, lambda: asyncio.create_task(client.close()))

        # вывод данных по подписке
        async for data in subs:
            print(data)
        print(f'\n\n\n== UnSubscribe ({subs.id}): MXSE.securities, TICKER="MXSE.TQBR.GAZP" and LANGUAGE="en"  ===')

        await client  # Ожидание пока клиент активен
        print(f'\n\n\n== End')


if __name__ == '__main__':
    asyncio.run(sample_websocket('wss://iss.moex.com/infocx/v3/websocket'))



from datetime import datetime, time, timedelta
from typing import Any, Iterable

from moexalgo.features.common import CommonMarket
from moexalgo.session import Session
from moexalgo.utils import DataFrame, result_adapter, normalize_data


class ExtraMarketMixin:
    """
    Миксин добавляющий методы: `trades`, `candles` обобщенному классу `Market`.
    """

    def trades(
        self: CommonMarket,
        tradeno: int | None = None,
        recno: int | None = None,
        *,
        native: bool = False,
    ) -> Iterable[dict[str, Any]] | DataFrame:
        """
        Возвращает последние сделки в обратной сортировке если `tradeno`/`recno` не заданно.
        Если задано сделки возвращаются от заданного `tradeno`/`recno`.

        Parameters
        ----------
        tradeno :
            Номер сделки до которого выдаются данные, для валютного и фондового рынка.
        recno :
            Номер порядка заключения сделок, для срочного рынка.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        key = "recno" if self.market in ("forts", "options") else "tradeno"
        options = dict()
        if tradeno is not None:
            options[key] = tradeno
        else:
            options["reversed"] = 1
        options["native"] = native
        return result_adapter(fetch_trades(self.engine, self.market, self.boardid, **options), native)

    def candles(self, native: bool = False) -> Iterable[dict[str, Any]] | DataFrame:
        """
        Метод возвращает две последние минутные свечи по всем инструментам рынка.

        Parameters
        ----------
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        data = list(reversed(list(self.trades(native=True))))
        if data:
            today = datetime.fromisoformat(data[-1]["systime"]).date()
            finish = datetime.combine(today, time.fromisoformat(data[-1]["tradetime"]))
            start = datetime.combine(today, time.fromisoformat(data[0]["tradetime"]))
            begin = datetime.fromtimestamp(int(finish.timestamp() / 60) * 60)
            end = begin + timedelta(minutes=1) - timedelta(microseconds=1)
            begin = begin - timedelta(minutes=1)
            while begin < start:
                tradeno = (int(data[0]["tradeno"]) - 3000) if data[0]["tradeno"] > 3000 else 0
                prev = [prev for prev in self.trades(tradeno, native=True) if prev["tradeno"] < data[0]["tradeno"]]
                data = prev + data
                start = datetime.combine(today, time.fromisoformat(data[0]["tradetime"]))
            trades = [
                trade
                for trade in [
                    dict(
                        ticker=r["ticker"],
                        tradetime=datetime.combine(today, time.fromisoformat(r["tradetime"])),
                        price=r["price"],
                        quantity=r.get("quantity"),
                        value=r.get("value"),
                    )
                    for r in data
                ]
                if begin <= trade["tradetime"] <= end
            ]
            data = make_candles(trades, begin, 60)
        return result_adapter(data, native)


def fetch_trades(engine, market: str, boardid: str, **options) -> Iterable[dict[str, Any]]:
    with Session() as client:
        return normalize_data(
            client.get_objects(
                f"engines/{engine}/markets/{market}/boards/{boardid}/trades", lambda data: data["trades"], **options
            )
        )


def make_candles(trades: list[dict[str, Any]], begin: datetime, interval: int) -> list[dict[str, Any]]:

    def make_candles_(ticker, data, end_):
        return dict(
            ticker=ticker,
            open=data[0]["price"],
            high=max(item["price"] for item in data),
            low=min(item["price"] for item in data),
            close=data[-1]["price"],
            volume=int(sum(item["quantity"] or 0 for item in data)),
            value=round(sum(item["value"] or 0 for item in data), 1),
            begin=end_ - timedelta(seconds=interval),
            end=(
                data[-1]["tradetime"]
                if end_ - timedelta(microseconds=1) >= datetime.now()
                else end_ - timedelta(microseconds=1)
            ).replace(microsecond=0),
        )

    trades_ = dict()
    for trade in sorted(trades, key=lambda t: (t["ticker"], t["tradetime"])):
        trades_.setdefault(trade["ticker"], []).append(trade)
    candles = []
    for ticker, trades in trades_.items():
        data = []
        end_ = begin + timedelta(seconds=interval)
        for trade in trades:
            if trade["tradetime"] < end_:
                data.append(trade)
            else:
                if data:
                    candles.append(make_candles_(ticker, data, end_))
                data = []
                end_ += timedelta(seconds=interval)
        if data:
            candles.append(make_candles_(ticker, data, end_))
    return candles



from datetime import date
from typing import Any, Iterable

from moexalgo.session import Session
from moexalgo.tools import resample
from moexalgo.utils import (
    DataFrame,
    result_adapter,
    normalize_data,
    CandlePeriod,
    normalize_period,
    prepare_from_till_dates,
    calc_offset_limit,
)


class CommonMarket:
    """
    Обобщенный класс `Market`

    Notes
    -----
    При использовании в среде Jupiter Notebook по умалчиванию данные возвращаются в виде `DataFrame`.
    Используйте в методах флаг `native` что бы отменить такое поведение.
    """

    def __init__(self, engine: str, market: str, boardid: str, features: dict[str, str]):
        self.features = features
        self.market = market
        self.boardid = boardid
        self.engine = engine

    def _get_path_for(self, feature):
        if path := self.features.get(feature):
            return path
        raise NotImplementedError(f"Methods of `{feature}` is not implemented for this endpoint.")

    def tickers(
        self,
        *fields: str,
        native: bool = False,
    ) -> Iterable[dict[str, Any]] | DataFrame:
        """
        Возвращает информация о всех инструментах рынка.

        Parameters
        ----------
        fields :
            Поля для отображения; "*" все поля
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        fields = () if "*" in fields else fields
        return result_adapter(
            normalize_data(fetch_securities(self.engine, self.market, self.boardid), *fields), native
        )

    def marketdata(
        self,
        *fields: str,
        native: bool = False,
    ) -> Iterable[dict[str, Any]] | DataFrame:
        """
        Возвращает статистическую информацию о всех инструментах рынка.

        Parameters
        ----------
        fields :
            Поля для отображения; "*" все поля
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        fields = () if "*" in fields else fields
        return result_adapter(
            normalize_data(fetch_securities(self.engine, self.market, self.boardid, "marketdata"), *fields), native
        )

    def __getattr__(self, name: str):
        if name in ("candles", "trades"):
            raise NotImplementedError(f"Method `{name}` not implemented for `{self.market}`")
        return self.__getattribute__(name)


def fetch_securities(engine, market: str, boardid: str, section: str = "securities") -> Iterable[dict[str, Any]]:
    with Session() as client:
        return client.get_objects(
            f"engines/{engine}/markets/{market}/boards/{boardid}/securities",
            lambda data: data[section],
        )


INFO_DEFAULT_FIELDS = (
    "title",
    "is_primary",
    "decimals",
    "is_traded",
    "market",
    "engine",
    "listed_from",
    "listed_till",
    "unit",
)


class CommonTicker:
    """
    Обобщенный класс `Ticker`

    Notes
    -----
    При использовании в среде Jupiter Notebook по умалчиванию данные возвращаются в виде `DataFrame`.
    Используйте в методах флаг `native` что бы отменить такое поведение.
    """

    def __init__(self, market: CommonMarket, boardid: str, secid: str, decimals: int, delisted: bool):
        self.market = market
        self.boardid = boardid
        self.secid = secid
        self.decimals = decimals
        self.delisted = delisted

    def info(
        self,
        *fields: str,
        native: bool = False,
    ) -> dict[str, Any] | DataFrame:
        """
        Информация об инструменте.

        Parameters
        ----------
        fields :
            Поля для отображения; "*" все поля
        native :
            Если флаг выставлен в `True` всегда возвращается словарь.
        """

        fields = (() if "*" in fields else fields) or INFO_DEFAULT_FIELDS

        def fetch():
            with Session() as client:
                data = normalize_data(client.get_objects(f"securities/{self.secid}", lambda data: data["boards"]))
                return [item for item in data if item["ticker"] == self.secid]

        return result_adapter(normalize_data(fetch(), *fields), native)

    def candles(
        self,
        start: str | date,
        end: str | date,
        period: str | int | CandlePeriod | None = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> Iterable[dict[str, Any]] | DataFrame:
        """
        Возвращает свечи инструмента по заданным параметрам.

        Parameters
        ----------
        start:
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end :
            Дата конца диапазона выдачи данных.
        period:
            Период свечи, возможны следующие строковые значения: '1min', '5min', '10min', '15min', '20min',
            '30min', '1h', '2h', '3h', '6h', '12h', '1D', '5D', '10D', '1W', '2W', '4W', '1M'; и числовые:
            1 (1 минута), 10 (10 минут), 60 (1 час), 24 (1 день), 7 (1 неделя), 31 (1 месяц).
        offset :
            Начальная позиция в последовательности записей, by default 0.
        latest :
            Включает режим выдачи последних `latest` записей в наборе, by default False.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        resample_to = None
        try:
            interval = normalize_period(period)
        except (ValueError, TypeError):
            interval, resample_to, period = resample.normalize_period(period)
        from_date, till_date = prepare_from_till_dates(start, end)
        options = {"from": from_date, "till": till_date, "interval": interval}
        offset, limit = calc_offset_limit(offset, 10_000)
        if latest:
            options["iss.reverse"] = True
            limit = 1
        options = dict(options, offset=offset, limit=limit)
        candles = fetch_section(
            self.market.engine,
            self.market.market,
            self.boardid,
            self.secid,
            "candles",
            **options,
        )
        return result_adapter(resample.candles(candles, resample_to) if resample_to is not None else candles, native)

    def trades(
        self,
        tradeno: int = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> Iterable[dict[str, Any]] | DataFrame:
        """
        Возвращает сделки за последний день или начиная с заданного `tradeno`.

        Parameters
        ----------
        tradeno :
            Номер сделки с которого выдаются данные, если не задано, то с начала дня.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        key = "recno" if self.market in ("forts", "options") else "tradeno"
        options = {key: tradeno}
        offset, limit = calc_offset_limit(offset, 10_000)
        if latest:
            options["reversed"] = 1
            limit = 1
        options = dict(options, offset=offset, limit=limit)
        return result_adapter(
            fetch_section(self.market.engine, self.market.market, self.boardid, self.secid, "trades", **options),
            native,
        )

    def orderbook(self, native: bool = False) -> Iterable[dict[str, Any]] | DataFrame:
        """
        Возвращает текущий стакан лучших цен.

        Parameters
        ----------
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        options = dict(offset=0, limit=-1)
        return result_adapter(
            fetch_section(self.market.engine, self.market.market, self.boardid, self.secid, "orderbook", **options),
            native,
        )


def fetch_section(
    engine: str,
    market: str,
    boardid: str,
    secid: str,
    section: str,
    *,
    limit: int,
    offset: int = 0,
    **options: Any,
) -> Iterable[dict[str, Any]]:
    start = offset
    with Session() as client:
        while True:
            options["start"] = start
            if data := normalize_data(
                client.get_objects(
                    f"engines/{engine}/markets/{market}/boards/{boardid}/securities/{secid}/{section}",
                    lambda data: data[section],
                    **options,
                )
            ):
                for item in data:
                    yield item
                    start += 1
                if (start - offset) < limit and limit > 0:
                    continue
            break



import sys
from datetime import date
from enum import Enum
from typing import Any, Iterable
from typing import Union, NamedTuple

import orjson as json  # noqa

pd = None
if "ipykernel" in sys.modules:
    try:
        import pandas as pd
        from pandas import DataFrame  # noqa
    except ImportError:
        pass
else:
    type DataFrame = Any


class CandlePeriod(Enum):
    """
    Возможные временные интервалы для свечей.
    """

    ONE_MINUTE = 1
    TEN_MINUTES = 10
    ONE_HOUR = 60
    ONE_DAY = 24
    ONE_WEEK = 7
    ONE_MONTH = 31


class Desc(NamedTuple):
    aliases: list[str]
    features: dict[str, str] | None = None


def normalize_period(period: CandlePeriod | int | str) -> int:
    """
    Нормализует значение периода во внутреннее представление ISS.
    """
    interval = CandlePeriod.TEN_MINUTES.value

    if isinstance(period, CandlePeriod):
        interval = period.value

    elif isinstance(period, int):
        if period == 1:
            interval = CandlePeriod.ONE_MINUTE.value
        elif period == 10:
            interval = CandlePeriod.TEN_MINUTES.value
        elif period == 60:
            interval = CandlePeriod.ONE_HOUR.value
        elif period == 24:
            interval = CandlePeriod.ONE_DAY.value
        elif period == 7:
            interval = CandlePeriod.ONE_WEEK.value
        elif period == 31:
            interval = CandlePeriod.ONE_MONTH.value
        else:
            raise ValueError("Некорректное значение параметра `period`")

    elif isinstance(period, str):
        if period == "1min":
            interval = CandlePeriod.ONE_MINUTE.value
        elif period == "10min":
            interval = CandlePeriod.TEN_MINUTES.value
        elif period in ("1h", "1H"):
            interval = CandlePeriod.ONE_HOUR.value
        elif period in ("1d", "1D"):
            interval = CandlePeriod.ONE_DAY.value
        elif period in ("1w", "1W"):
            interval = CandlePeriod.ONE_WEEK.value
        elif period in ("1m", "1M"):
            interval = CandlePeriod.ONE_MONTH.value
        else:
            raise ValueError("Некорректное значение параметра `period`")

    elif period is None:
        pass
    else:
        raise TypeError("Неверный тип для `period`")

    return interval


def prepare_from_till_dates(from_date: Union[str, date] = None, till_date: Union[str, date] = None) -> tuple[str, str]:
    """
    Подготовка дат начала и окончания.

    Parameters
    ----------
    from_date :
        Дата начала, by default None.
    till_date :
        Дата окончания, by default None.

    Returns
    -------
    return : tuple[str, str]
        Кортеж с датами начала и окончания.
    """
    if (from_date is None) or (till_date is None):
        raise ValueError("Оба параметра `from_date` и `till_date` должны быть заданы")

    from_date = date.fromisoformat(from_date) if isinstance(from_date, str) else (from_date or date.today())
    if not till_date:
        till_date = from_date
    elif isinstance(till_date, str):
        till_date = date.today() if till_date == "today" else date.fromisoformat(till_date)

    if from_date > till_date:
        raise ValueError("Параметр `from_date` должен быть меньше или равный `till_date`")

    return from_date.isoformat(), till_date.isoformat()


def calc_offset_limit(
    offset: int = None,
    limit: int = None,
    min_limit: int = 1,
    standart_limit: int = 10_000,
    max_limit: int = 50_000,
    min_offset: int = 0,
) -> tuple[int, int]:
    """
    Вычисление смещения и лимита.

    Parameters
    ----------
    offset : int, optional
        Смещение относительно начала, by default None.
    limit : int, optional
        Лимит данных, by default None.

    Returns
    -------
    return : tuple[int, int]
        Кортеж со смещением и лимитом.
    """
    offset = offset or min_offset
    if limit != -1:
        limit = limit or standart_limit

        limit = min_limit if limit < min_limit else max_limit if limit > max_limit else limit
        offset = min_offset if offset < min_offset else max_limit - 1 if offset >= max_limit else offset

    return offset, limit


def normalize_data(data: Iterable[dict[str, Any]], *fields: str | tuple[str, str]) -> Iterable[dict[str, Any]]:
    names = {"secid": "ticker", "boardid": "board"}
    names.update(dict((f, f) if isinstance(f, str) else (f[0], f[1]) for f in fields))

    keys = [names.get(name.lower(), name.lower()) for name in data["columns"]]
    rows = data["data"]
    data = [dict(zip(keys, row)) for row in rows]
    if fields:
        keys = tuple(names.values())
        return [dict((k, v) for k, v in row.items() if k in keys) for row in data]
    return data


def result_adapter(result: dict | Iterable[dict], native: bool) -> dict | Iterable[dict] | DataFrame | None:
    """Преобразует результат в `pandas.DataFrame` если применимо."""
    use_dataframe = not native and pd is not None
    if use_dataframe:
        if result is None:
            return None
        result = pd.DataFrame([result] if isinstance(result, dict) else list(result))
    return result



"""
Пример для вебсокетов, как делать несколько подписок на одном соединении
"""

import asyncio

from moexalgo.beta import issplus


async def gen_subscription(client, destination, selector):
    subscription = await client.subscribe(destination, selector.format(ticker="MOEX"))
    # Отмена подписки через 45 сек
    asyncio.get_running_loop().call_later(45.0, lambda: asyncio.create_task(client.unsubscribe(subscription.id)))
    # Получение данных
    columns = []
    async for data in subscription:
        print(selector)
        if not columns:
            columns.extend(data["columns"])
            print("\t".join(columns))
        for rec in data["data"]:
            print("\t".join(map(str, [round(*cell) if isinstance(cell, list) else cell for cell in rec])))


async def main():
    url = "wss://iss.moex.com/infocx/v3/websocket"
    destination = "MXSE.orderbooks"
    selector = 'TICKER="MXSE.TQBR.{ticker}" and LANGUAGE="en"'
    tickers = ("MOEX", "GAZP")

    async with issplus.connect(url, issplus.Credentials("DEMO", "guest", "guest")) as client:
        tasks = [
            asyncio.create_task(gen_subscription(client, destination, selector.format(ticker=ticker)))
            for ticker in tickers
        ]

        await asyncio.gather(*tasks)


if __name__ == "__main__":
    asyncio.run(main())



import os

import pytest

from moexalgo import Ticker, session

session.TOKEN = os.environ["APIKEY"]


def test_market():
    moex = Ticker("MOEX")

    assert len(moex.info())

    it = moex.tradestats(start="2024-01-05", end="2024-02-05")
    assert isinstance(next(it), object)
    data = list(it)
    assert len(data) > 500


if __name__ == "__main__":
    pytest.main()



from typing import Iterable, Any

from moexalgo.features.algopack import AlgopackMarketMixin, AlgopackTickerMixin
from moexalgo.features.common import CommonMarket, CommonTicker
from moexalgo.features.extramarket import ExtraMarketMixin
from moexalgo.utils import DataFrame

SELT_TICKERS_DEFAULT_FIELDS = (
    "shortname",  # Краткое наименование
    "lotsize",  # Размер лота
    "decimals",  # Количество знаков после запятой
    "minstep",  # Минимальный шаг цены
    "secname",
)

SELT_MARKET_DEFAULT_FIELDS = (
    "bid",  # Лучшая цена покупки
    "offer",  # Лучшая цена продажи
    "biddeptht",  # Глубина стакана покупки
    "offerdeptht",  # Глубина стакана продажи
    "open",  # Цена открытия
    "high",  # Максимальная цена
    "low",  # Минимальная цена
    "last",  # Цена последней сделки
    "waprice",  # Средневзвешенная цена
    "lasttoprevprice",  # Изменение цены последней сделки к предыдущей
    "numtrades",  # Количество сделок
    "voltoday",  # Объем сделок за день
    "valtoday",  # Объем сделок за день в валюте
    "valtoday_usd",  # Объем сделок за день в долларах
    "updatetime",  # Время обновления
    "systime",  # Время системы
)


class Market(CommonMarket, ExtraMarketMixin, AlgopackMarketMixin):
    """
    Обобщенный класс `Market` для валютного рынка.
    """

    def __init__(self, *args, **kwargs):
        super().__init__("currency", *args, **kwargs)

    def tickers(self, *fields: str, native: bool = False) -> Iterable[dict[str, Any]] | DataFrame:
        match self.market:
            case "selt":
                fields = () if "*" in fields else (fields or SELT_TICKERS_DEFAULT_FIELDS)
            case _:
                fields = () if "*" in fields else fields
        return super().tickers(*fields, native=native)

    def marketdata(self, *fields: str, native: bool = False) -> Iterable[dict[str, Any]] | DataFrame:
        match self.market:
            case "selt":
                fields = () if "*" in fields else (fields or SELT_MARKET_DEFAULT_FIELDS)
            case _:
                fields = () if "*" in fields else fields
        return super().marketdata(*fields, native=native)


class Ticker(CommonTicker, AlgopackTickerMixin):
    """
    Обобщенный класс `Ticker` для валютного рынка.
    """



import typing as t
from datetime import date

from moexalgo.features.common import CommonTicker, CommonMarket
from moexalgo.session import Session
from moexalgo.utils import (
    DataFrame,
    result_adapter,
    prepare_from_till_dates,
    calc_offset_limit,
    normalize_data,
)


class AlgopackMarketMixin:
    """
    Миксин добавляющий методы Algopack обобщенному классу `Market`.
    """

    def tradestats(
        self: CommonMarket,
        *,
        date: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает метрики `TradeStat` по заданным параметрам.

        Parameters
        ----------
        date :
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self._get_path_for("algopack")
        date, _ = prepare_from_till_dates(date, date)
        offset, limit = calc_offset_limit(offset, 50_000)
        options = dict(date=date, offset=offset, limit=limit)
        if latest:
            options["latest"] = 1
        return result_adapter(fetch_algopack(path, "tradestats", **options), native)

    def orderstats(
        self: CommonMarket,
        *,
        date: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает метрики `OrderStat` по заданным параметрам.

        Parameters
        ----------
        date :
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self._get_path_for("algopack")
        date, _ = prepare_from_till_dates(date, date)
        offset, limit = calc_offset_limit(offset, 50_000)
        options = dict(date=date, offset=offset, limit=limit)
        if latest:
            options["latest"] = 1
        return result_adapter(fetch_algopack(path, "orderstats", **options), native)

    def obstats(
        self: CommonMarket,
        *,
        date: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает метрики `ObStat` по заданным параметрам.

        Parameters
        ----------
        date :
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self._get_path_for("algopack")
        date, _ = prepare_from_till_dates(date, date)
        offset, limit = calc_offset_limit(offset, 50_000)
        options = dict(date=date, offset=offset, limit=limit)
        if latest:
            options["latest"] = 1
        return result_adapter(fetch_algopack(path, "obstats", **options), native)

    def alerts(
        self: CommonMarket,
        *,
        date: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает MegaAlert (оповещение об аномальной рыночной активности) по заданным параметрам.

        Parameters
        ----------
        date :
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self._get_path_for("algopack")
        date, _ = prepare_from_till_dates(date, date)
        offset, limit = calc_offset_limit(offset, 50_000)
        options = dict(date=date, offset=offset, limit=limit)
        if latest:
            options["latest"] = 1
        return result_adapter(fetch_algopack(path, "alerts", **options), native)

    def hi2(
        self: CommonMarket,
        *,
        date: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает метрики `Hi2` (индекс рыночной концентрации) по заданным параметрам.

        Parameters
        ----------
        date :
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self._get_path_for("algopack")
        date, _ = prepare_from_till_dates(date, date)
        offset, limit = calc_offset_limit(offset, 50_000)
        options = dict(date=date, offset=offset, limit=limit)
        if latest:
            options["latest"] = 1
        return result_adapter(fetch_algopack(path, "hi2", **options), native)


class AlgopackTickerMixin:
    """
    Миксин добавляющий методы Algopack обобщенному классу `Ticker`.
    """

    def tradestats(
        self: CommonTicker,
        *,
        start: str | date = None,
        end: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает метрики `TradeStat` по заданным параметрам.

        Parameters
        ----------
        start :
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end :
            Дата конца диапазона выдачи данных.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self.market._get_path_for("algopack")
        from_date, till_date = prepare_from_till_dates(start, end)
        offset, limit = calc_offset_limit(offset, 10_000)
        options = dict(**{"from": from_date, "till": till_date}, offset=offset, limit=limit)
        if latest:
            options["latest"] = 1
        return result_adapter(fetch_algopack(path, "tradestats", self.secid, **options), native)

    def orderstats(
        self: CommonTicker,
        *,
        start: str | date = None,
        end: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает метрики `OrderStat` по заданным параметрам.

        Parameters
        ----------
        start :
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end :
            Дата конца диапазона выдачи данных.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self.market._get_path_for("algopack")
        from_date, till_date = prepare_from_till_dates(start, end)
        offset, limit = calc_offset_limit(offset, 10_000)
        options = dict(**{"from": from_date, "till": till_date}, offset=offset, limit=limit)
        if latest:
            options["latest"] = 1
        return result_adapter(fetch_algopack(path, "orderstats", self.secid, **options), native)

    def obstats(
        self: CommonTicker,
        *,
        start: str | date = None,
        end: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает метрики `ObStat` по заданным параметрам.

        Parameters
        ----------
        start :
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end :
            Дата конца диапазона выдачи данных.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self.market._get_path_for("algopack")
        from_date, till_date = prepare_from_till_dates(start, end)
        offset, limit = calc_offset_limit(offset, 10_000)
        options = dict(**{"from": from_date, "till": till_date}, offset=offset, limit=limit)
        if latest:
            options["latest"] = 1
        return result_adapter(fetch_algopack(path, "obstats", self.secid, **options), native)

    def alerts(
        self: CommonTicker,
        *,
        start: str | date = None,
        end: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает MegaAlert (оповещение об аномальной рыночной активности) по заданным параметрам.

        Parameters
        ----------
        start :
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end :
            Дата конца диапазона выдачи данных.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self.market._get_path_for("algopack")
        from_date, till_date = prepare_from_till_dates(start, end)
        offset, limit = calc_offset_limit(offset, 10_000)
        options = dict(**{"from": from_date, "till": till_date}, offset=offset, limit=limit)
        if latest:
            options["latest"] = 1
        return result_adapter(fetch_algopack(path, "alerts", self.secid, **options), native)

    def hi2(
        self: CommonTicker,
        *,
        start: str | date = None,
        end: str | date = None,
        latest: bool = None,
        offset: int = None,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает метрики `Hi2` (индекс рыночной концентрации) по заданным параметрам.

        Parameters
        ----------
        start :
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end :
            Дата конца диапазона выдачи данных.
        latest :
            Включает режим выдачи последних записей в наборе.
        offset :
            Начальная позиция в последовательности записей.
        native :
            Если флаг выставлен в `True` всегда возвращается итератор словарей.
        """
        path = self.market._get_path_for("algopack")
        from_date, till_date = prepare_from_till_dates(start, end)
        offset, limit = calc_offset_limit(offset, 10_000)
        options = dict(**{"from": from_date, "till": till_date}, offset=offset, limit=limit)
        if latest:
            options["latest"] = 1
        return result_adapter(fetch_algopack(path, "hi2", self.secid, **options), native)


def fetch_algopack(
    path: str,
    metrics: str,
    secid: str = None,
    limit: int = 10_000,
    offset: int = 0,
    **options: t.Any,
):
    with Session() as client:
        start = offset
        while True:
            options["start"] = start
            if data := normalize_data(
                client.get_objects(
                    f"{path}/{metrics}/{secid}" if secid else f"{path}/{metrics}", lambda data: data["data"], **options
                )
            ):
                for item in data:
                    yield item
                    start += 1
                if (start - offset) < limit and limit > 0:
                    continue
            break



import csv
import typing as t
import zipfile
from dataclasses import dataclass
from datetime import datetime
from io import BytesIO, TextIOWrapper

from moexalgo.utils import pd


@dataclass
class FileInfo:
    id: int
    dataset: str
    file_size: int
    filename: str
    uploaded_at: datetime
    year_month: str


class FileList(t.List[FileInfo]):

    @property
    def df(self) -> pd.DataFrame:
        """Dataframe representation of this object."""
        return pd.DataFrame(self)

    def _repr_html_(self):
        return self.df._repr_html_()


class FileContent(t.List[str]):

    @property
    def df(self) -> pd.DataFrame:
        """Dataframe representation of this object."""
        return pd.DataFrame(dict(csv=self))

    def _repr_html_(self):
        return self.df._repr_html_()


class FileData(t.List[t.Dict]):

    @property
    def df(self) -> pd.DataFrame:
        """Dataframe representation of this object."""
        return pd.DataFrame(self)

    def _repr_html_(self):
        return self.df._repr_html_()


@dataclass
class DownloadInfo:
    download_url: str
    file_id: int
    filename: str
    success: bool


class File:

    def __init__(self, content: bytes):
        self.__data = None
        self._stream = BytesIO(content)
        with zipfile.ZipFile(self._stream, "r") as zip_file:
            self.__content = [name for name in zip_file.namelist() if name.endswith(".csv")]
        self._stream.seek(0)

    def __iter__(self):
        return iter(self._data)

    @property
    def df(self) -> pd.DataFrame:
        """Dataframe representation of this object."""
        return self._data.df

    @property
    def _data(self) -> FileData:
        if self.__data is None:
            self.__data = self._select(self._content[0])
        return self.__data

    @property
    def _content(self) -> FileContent:
        if not self.__content:
            raise ValueError("No csv files found in zip file")
        return FileContent(self.__content)

    def _select(self, csv_file: str) -> FileData:
        with zipfile.ZipFile(self._stream, "r") as zip_file:
            with zip_file.open(csv_file) as file:
                text_file = TextIOWrapper(file, encoding="utf-8")
                reader = csv.DictReader(text_file)
                return FileData(row for row in reader)
